/**
 * Copyright © 2019, Oracle and/or its affiliates. All rights reserved.
 *
 * Licensed under the Universal Permissive License v 1.0 as shown at
 * http://oss.oracle.com/licenses/upl
 *
 */
#define _GNU_SOURCE
#include "putils.h"
#include "malloc.h"
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

FILE* do_creat(const char *fullname) {
    int fd = open(fullname, O_CREAT | O_WRONLY, S_IWUSR);
    if (fd == -1) {
        fprintf(stderr, "Trying to open %s:\n", fullname);
        perror("open");
        abort();
    }
    assert(fd >= 0);
    FILE *f = fdopen(fd, "w");
    if (!f) perror("fdopen");
    assert(f);
    return f;
}

FILE* do_open_for_read(const char *fullname) {
    int fd = open(fullname, O_RDONLY);
    if (fd == -1) {
        fprintf(stderr, "Trying to open %s:\n", fullname);
        perror("open");
        abort();
    }
    assert(fd >= 0);
    FILE *f = fdopen(fd, "r");
    if (!f) perror("fdopen");
    assert(f);
    return f;
}


void do_fclose(FILE *f) {
    int r = fclose(f);
    if (r != 0) perror("fclose");
    assert(r == 0);
}

void do_write(FILE *f, const char *bytes, size_t size) {
    size_t r = fwrite(bytes, 1, size, f);
    if (r != size) perror("fwrite");
    assert(r == size);
}

size_t do_read(FILE *f, char *bytes, size_t size) {
    // We might be reading from a socket or a pipe, so keep reading till we get all the bytes.
    size_t r = fread(bytes, 1, size, f);
    if (r == 0) {
        if (feof(f)) return 0;
        assert(ferror(f));
        fprintf(stderr, "do_read got error %d\n", ferror(f));
        assert(0);
    }
    if (r < size) {
        return r + do_read(f, bytes+r, size-r);
    } else {
        return r;
    }
}

void do_mkdir(const char *fullname) {
    if (0) fprintf(stderr, "mkdir(\"%s\");\n", fullname);
    int r = mkdir(fullname, S_IRUSR|S_IWUSR|S_IXUSR);
    if (r != 0) {
        if (errno == EEXIST) {
            return;
        }
        // We ensure that the parent exists before we call do_mkdir,
        // so ENOENT is a real error.
        fprintf(stderr, "Trying to mkdir %s\n", fullname);
        perror("mkdir");
        abort();
    }
    assert(r == 0);
}

void do_mkfifo(const char *fullname, mode_t mode) {
    int r = mkfifo(fullname, mode);
    if (r) {
        fprintf(stderr, "Trying to mkfifo(%s, %d)\n", fullname, mode);
        perror("mkfifo");
        abort();
    }


}

void do_link(const char *linkto, const char *fullname) {
    int r = link(linkto, fullname);
    if (r) {
        fprintf(stderr, "Trying to link \"%s\" --> \"%s\"\n", fullname, linkto);
        perror("link");
        abort();
    }
}


void do_symlink(const char *contents, const char *fullname) {
    int r = symlink(contents, fullname);
    r = r; // unused
}

void do_unlink(const char *fullname) {
    int r = unlink(fullname);
    assert(r == 0);
}
void do_rmdir(const char *fullname) {
    int r = rmdir(fullname);
    assert(r == 0);
}


char *pathcat(const char *a, const char *b) {
    char *result = NULL;
    int r = asprintf(&result, "%s/%s", a, b);
    assert(r > 0);
    return result;
}

size_t parse_number_or_help(const char *numstring) {
    errno = 0;
    char *end;
    size_t result = strtoul(numstring, &end, 0);
    if (0
        || errno != 0
        || *end != 0)
    {
        fprintf(stderr, "Trying to parse %s as a number\n", numstring);
        help(1);
    }
    return result;
}

void
maybe_error(int r, const char *fun, const char *fname, const char *srcfile, int line) {
    if (r != 0) {
        fprintf(stderr, "%s:%d Could not %s(\"%s\").  errno=%d (%s)\n",
                srcfile, line, fun, fname, errno, strerror(errno));
        exit(1);
    }
}

void
maybe_error2(int r, const char *fun, const char *fname, const char *fname2, const char *srcfile, int line) {
    if (r != 0) {
        fprintf(stderr, "%s:%d Could not %s(\"%s\", \"%s\").  errno=%d (%s)\n",
                srcfile, line,
                fun, fname, fname2, errno, strerror(errno));
        exit(1);
    }
}
